/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "unlink.c" - Delete files by name and unlink allocation chains *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

//#define DEBUG

#ifdef DEBUG
#endif


/* Unlinks (marks as free) all the clusters of the chain starting from     */
/* cluster FromCluster (that must be part of a valid chain) until the EOC. */
/* Called by fat_unlink and truncate_or_extend (write.c).                  */
int unlink_clusters(tFatVolume *V, DWORD FromCluster)
{
  int    k, Res;
  DWORD  Cluster = FromCluster;
  DWORD  Next    = 0;

  switch (V->FatType)
  {
    case FATTYPE_FAT12:
      do
      {
        #ifdef DEBUG
        printf("Unlinking cluster %lu\n",Cluster);
        #endif
        if ((Res = fat12_read_entry(V, Cluster, 0, &Next))) return Res;
        /* Update every FAT in the volume */
        for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
          if ((Res = fat12_write_entry(V, Cluster, k, 0))) return Res;
        Cluster = Next;
      }
      while (!(FAT12_EOC(Next)));
      return 0;

    case FATTYPE_FAT16:
      do
      {
        if ((Res = fat16_read_entry(V, Cluster, 0, &Next))) return Res;
        /* Update every FAT in the volume */
        for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
          if ((Res = fat16_write_entry(V, Cluster, k, 0))) return Res;
        Cluster = Next;
      }
      while (!(FAT16_EOC(Next)));
      return 0;

    case FATTYPE_FAT32:
      do
      {
        if ((Res = fat32_read_entry(V, Cluster, 0, &Next))) return Res;
        /* Update every FAT in the volume */
        for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
          if ((Res = fat32_write_entry(V, Cluster, k, 0))) return Res;
        Cluster = Next;
      }
      while (!(FAT32_EOC(Next)));
      return 0;

    default: /* should never arrive here */
             return FAT_RET_CRITICAL_ERROR;
  }
}


/* Deletes the specified file(s) according to flags.                    */
/* Returns 0 on success, or a negative error code on failure.           */
/* This is a public driver function, and is called by fat_rmdir (dir.c) */
int fat_unlink(void *P, char *FileName, DWORD Flags, int UseLfn)
{
  tFatVolume *V = (tFatVolume *) P;
  tFatFile   *Dir;
  tFatFind    D;
  int         Res, Fd;
  char        Path[FD32_MAX_LFN_PATH_LENGTH];
  char        Name[FD32_MAX_LFN_LENGTH];
  int         NoneDeleted = 1; /* Check if something has been deleted   */
  DWORD       SaveDirPos;      /* Save the dir position while searching */
/*
  TO DO: This will go in the INT 21h handler.
  unsigned char AllowableAttributes = FD32_ATTR_READONLY | FD32_ATTR_ARCHIVE |
                                      FD32_ATTR_HIDDEN   | FD32_ATTR_SYSTEM;
  if (RequiredAttributes == FD32_ATTR_DIRECTORY)
    AllowableAttributes |= FD32_ATTR_DIRECTORY;
*/
  
  split_path(FileName, Path, Name);
  #ifdef DEBUG
  printf("Unlinking file %s in directory %s\n", Name, Path);
  #endif
  Fd = fat_open(V, Path, FD32_OPEN_ACCESS_READWRITE | FAT_OPEN_DIRECTORY,
                FD32_ATTR_NONE, 0, NULL, UseLfn);
  if (Fd < 0) return Fd;
  Dir = &V->Files[Fd];

  /* Repeat until no more matches are found */
  for (;;)
  {
    Res = find(V, Fd, Name, Flags, &D, UseLfn);
    SaveDirPos = Dir->TargetPos;

    /* Check if error (usually file not found or read only file) */
    switch (Res)
    {
      case 0                        : break;
      case FD32_ERROR_NO_MORE_FILES : fat_close(V, Fd);
                                      if (NoneDeleted) return Res;
                                                  else return 0;
      default                       : fat_close(V, Fd);
                                      return Res;
    }
    if (D.SfnEntry.Attr & FD32_ATTR_READONLY) return FD32_ERROR_ACCESS_DENIED;

    Res = close_shared_file(V, (((unsigned long)D.SfnEntry.FstClusHI) << 16) + D.SfnEntry.FstClusLO);
    if (Res) return Res;
    
    /* If the file is not zero bytes long we unlink its cluster chain */
    if ((((unsigned long)D.SfnEntry.FstClusHI) << 16) + D.SfnEntry.FstClusLO != 0)
      unlink_clusters(V, (((unsigned long)D.SfnEntry.FstClusHI) << 16) + D.SfnEntry.FstClusLO);

    Res = free_dir_entries(V, Fd, D.EntryOffset, D.LfnEntries);
    if (Res) return Res;
    NoneDeleted = 0;
    Dir->TargetPos = SaveDirPos;
  }
}
