/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "write.c" - Write a block of data into an open file            *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

//#define DEBUG


/* Returns nonzero if the block of size Size to write into the file is */
/* too large to fit into the free volume space.                        */
/* NOTE: Doens't work if the file is a directory because FileSize==0.  */
/* Called by fat_write.                                                */
static /*inline*/ int block_is_too_large(tFatVolume *V, int Fd, DWORD Size)
{
  tFatFile *F                 = &V->Files[Fd];
  DWORD     FinalOccupation   = clusters_amount(F->TargetPos + Size, &V->Bpb);
  DWORD     InitialOccupation = clusters_amount(F->DirEntry.FileSize, &V->Bpb);

  /* Always return success for directories */
  if (F->DirEntry.Attr & FD32_ATTR_DIRECTORY) return 0;

  #ifdef DEBUG
  printf("Initial occupation : %lu\n",InitialOccupation);
  printf("Final occupation   : %lu\n",FinalOccupation);
  printf("Free clusters      : %lu\n",V->FSI_Free_Count);
  #endif

  /* If the file is not longer than before we are ok */
  if (FinalOccupation <= InitialOccupation) return 0;
  /* If the file surplus fits into the disk's free clusters we are ok */
  if (FinalOccupation - InitialOccupation <= V->FSI_Free_Count) return 0;
  /* Otherwise the block is too large */
  return 1;
}


/* Truncates or extends a file to the position TargetPos, as required */
/* by a write of zero bytes.                                          */
/* Returns 0 on success, or a negative error code on failure.         */
/* Called by fat_write.                                               */
static int truncate_or_extend(tFatVolume *V, int Fd)
{
  int       Res;
  DWORD     k;
  tFatFile *F = &V->Files[Fd];

  /* TargetPos == 0 is Truncate to zero length, unlinking */
  /* all file's clusters, including the first.            */
  if (F->TargetPos == 0)
  {
    /* If the file is already long zero we do nothing */
    if (F->FirstCluster == 0) return 0;
    unlink_clusters(V, F->FirstCluster);
    F->FirstCluster      = 0;
    F->DirEntry.FileSize = 0;
    return 0;
  }
  /* TragetPos > 0 is Truncate/Extend at TargetPos position _excluded_,  */
  /* so we go to position TargetPos-1, since a writing seek to TargetPos */
  /* would allocate one byte more.                                       */
  // [KG] update: a writing seek does not allocate 1 extra byte anymore
  //  F->TargetPos--;
  Res = move_to_targetpos(V, Fd, 0);
  //  F->TargetPos++;
  if (Res) return Res;
  /* Now if there are clusters beyond here we unlink */
  /* them, and mark the current cluster with EOC.    */
  #ifdef DEBUG
  printf("truncate_or_extend: Now unlinking...\n");
  #endif
  switch (V->FatType)
  {
    case FATTYPE_FAT12 :
      Res = fat12_read_entry(V, F->Cluster, 0, &k); /* the next cluster */
      if (Res) return Res;
      if (!(FAT12_EOC(k))) unlink_clusters(V, k);
      for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
      {
        Res = fat12_write_entry(V, F->Cluster, k, 0x0FFF);
        if (Res) return Res;
      }
      break;
    case FATTYPE_FAT16 :
      Res = fat16_read_entry(V, F->Cluster, 0, &k); /* the next cluster */
      if (Res) return Res;
      if (!(FAT16_EOC(k))) unlink_clusters(V, k);
      for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
      {
        Res = fat16_write_entry(V, F->Cluster, k, 0xFFFF);
        if (Res) return Res;
      }
      break;
    case FATTYPE_FAT32 :
      Res = fat32_read_entry(V, F->Cluster, 0, &k); /* the next cluster */
      if (Res) return Res;
      if (!(FAT32_EOC(k))) unlink_clusters(V, k);
      for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
      {
        Res = fat32_write_entry(V, F->Cluster, k, 0x0FFFFFFF);
        if (Res) return Res;
      }
      break;
  }
  /* And finally set the new file size to where we've sought */
  F->DirEntry.FileSize = F->TargetPos;

  #ifdef DEBUG
  printf("File successfully truncated/extended at offset %lu\n",
          F->DirEntry.FileSize);
  #endif
  return 0;
}


/* Writes Size bytes from Buffer into a file.                       */
/* Returns 0 on success, or a negative error code on failure.       */
/* Puts the actual number of bytes written in Result (if not NULL). */
/* This is a public driver function.                                */
int fat_write(void *P, int Fd, BYTE *Buffer, DWORD Size, DWORD *Result)
{
  DWORD       k;
  SDWORD      Res;
  int         NumBuf;
  tFatVolume *V = (tFatVolume *) P;
  tFatFile   *F = validate_fd(V, Fd);
unsigned int SectorOffset;

  /* If some error will happen, we'll return 0 bytes written */
  if (Result) *Result = 0;
  if (F == NULL) return FD32_ERROR_INVALID_HANDLE;

  /* Check if writing into file is allowed */
  if (((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_WRITE)
   && ((F->Mode & FD32_OPEN_ACCESS_MASK) != FD32_OPEN_ACCESS_READWRITE))
    return FD32_ERROR_ACCESS_DENIED;

  /* Check if the file will fits into disk space after writing this block */
  if (block_is_too_large(V, Fd, Size)) return FD32_ERROR_DISK_FULL;

  /* Tell to actual_seek() that we are going to write */
  F->Mode |= FAT_FMODE_DO_WRITE;

  /* Special case: if Size is zero, the file has to be truncated/extended */
  /* to the current target position, if it's not a FAT12/FAT16 root dir.  */
  if (Size == 0)
  {
    if (F->Mode & FAT_FMODE_ROOT) return FD32_ERROR_ACCESS_DENIED;
                             else return truncate_or_extend(V, Fd);
  }
  /* Now really write */
  for (k=0; k<Size; k++)
  {
    Res = move_to_targetpos(V, Fd, Size-k);
    if (Result) *Result = k;
    /* This is possible only if the file is a FAT12/FAT16 */
    /* root directory with no more room.                  */
    if (Res == FAT_RET_EOF) return FAT_RET_DIRECTORY_FULL;
    else if (Res) return Res;
    /* Then we load the sector we sought, change the byte */
    /* we want and write back the sector.                 */
    if (F->Mode & FAT_FMODE_ROOT) {
      NumBuf = read_from_buffer(V, F->SectorInCluster +
                                first_root_dir_sector(&V->Bpb), &SectorOffset);
      }
    else
      NumBuf = read_from_buffer(V, F->SectorInCluster +
                                first_sector_of_cluster(F->Cluster, V), &SectorOffset);
    if (NumBuf < 0) return NumBuf;

//[KG] SPEEDUP: transfer more than one byte, if possible...
//Noch zu bertragende Bytes: Size-k
//Im Puffer Verfgbare Bytes: 512-(F->ByteInSector)
    unsigned long TransferSize, LeftBytesInBuffer;
    TransferSize = Size-k;
    LeftBytesInBuffer = 512-F->ByteInSector; //(512*V->SectorsInBuffer)-(F->ByteInSector+SectorOffset);
    if( TransferSize > LeftBytesInBuffer ) {
      TransferSize = LeftBytesInBuffer;
    }
    if (Buffer == 0)
      memset(&(V->Buffer[NumBuf].Data[F->ByteInSector+SectorOffset]), 0, TransferSize);
    else
      memcpy(&(V->Buffer[NumBuf].Data[F->ByteInSector+SectorOffset]), &Buffer[k], TransferSize);


//    V->Buffer[NumBuf].Data[F->ByteInSector+SectorOffset] = Buffer ? Buffer[k] : 0;
    V->Buffer[NumBuf].Flags |= FAT_BUFFER_DIRTY;
    /* Finally we advance the file position */

    F->TargetPos += TransferSize;
    k += TransferSize-1;
  }
  Res = move_to_targetpos(V, Fd, 0);

  /* Successful exit, whole Buffer written */
  if (Result) *Result = k;
  if (!(F->Mode & FAT_OPEN_DIRECTORY))
  {
    update_date_and_time(&F->DirEntry.WrtTime,
                         &F->DirEntry.WrtDate, NULL);
    F->DirEntry.LstAccDate = F->DirEntry.WrtDate;
  }
  if (F->Mode & FD32_OPEN_ALWAYS_COMMIT)
    if ((Res = fat_fflush(V, Fd))) return Res;
  return 0;
}
