/*  main.cpp
    IPC@Chip extdisk driver

    28.08.2001: tk, command line parsing rewritten, eliminated the
                    diskio function, LBA<>CHS conversion functions,
                    etc.. the partition size is now determined from
                    the CHS values (no more LBA! some fdisk's write bogus
                    values - see HaleLandis "how it all works"), and merged
                    in some enhancements made by thorsten godau:
                tg, modified to display model, revision and serialnumber,
                    modified to display more infos about partition type.
    27.08.2001: tk, implemented command line switches, and options
    26.08.2001: tk, all physical harddisk accesses are changed to CHS mode,
                    because many old harddisks <528 MB don't support LBA.
    25.08.2001: tk, sourcecode taken from Beck IDE100 example.

    tk: thomas.kindler@gmx.de
    tg: thorsten.godau@tz-mikroelektronik.de

    Copyright (c)2001 by Thomas Kindler, Thorsten Godau

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version. Read the
    full License at http://www.gnu.org/copyleft for more details.
*/

// include files
//
#include <ctype.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include "getopt.h"
#include "lbaconv.h"
#include "ide.h"
#include "hwapi.h"
#include "rtos.h"
#include "amd186.h"

#include <mem.h>
#include <string.h>

#include "fat1.h"

/* SUPPORT.C - Small support procedures */
#include "support.cpp"
/* BUFFERS.C - Bufferized I/O support procedures */
#include "buffers.cpp"
/* DATETIME.C - Files' date and time procedures */
#include "datetime.cpp"
/* DIR.C - Directory management services */
#include "dir.cpp"
/* DIRENTRY.C - Directory entry management procedures */
#include "direntry.cpp"
/* FAT12.C - FAT12 table access procedures */
#include "fat12.cpp"
/* FAT16.C - FAT16 table access procedures */
#include "fat16.cpp"
/* FAT32.C - FAT32 table access procedures */
#include "fat32.cpp"
/* INIT.C - Mount a FAT volume initializing all its data structures */
#include "init.cpp"
/* FIND.C - File find procedures */
#include "find.cpp"
/* LFN.C - Long file names support procedures */
#include "lfn.cpp"
/* OPEN.C - File opening and creation procedures */
#include "open.cpp"
/* READ.C - Read a block of data from a file */
#include "read.cpp"
/* RENAME.C - Rename files and directories */
#include "rename.cpp"
/* SEEK.C - Move file position to the target position specified by lseek */
#include "seek.cpp"
/* SHARE.C - Sharing conflict detection */
#include "share.cpp"
/* SHORTNAM.C - Support code for short file names (8.3 format) */
#include "shortnam.cpp"
/* UNLINK.C - File and clusters unlinking procedures */
#include "unlink.cpp"
/* WRITE.C - Write a block of data into a file truncating or extending it */
#include "write.cpp"
#include "unicode.cpp"
#include "wildcard.cpp"
#include "int.cpp"


#define  MAX_PARTITIONS  16

typedef struct {
  char  bootData[446];
  struct {
    char     active;
    char     s_head;
    unsigned s_cylsec;
    char     type;
    char     e_head;
    unsigned e_cylsec;
    unsigned long  offset;
    unsigned long  size;
  } part[4];
  unsigned magic;
} PartSector;


//typedef struct {
//  int            active, type;
//  unsigned long  start, size;
//} Partition;


// Command line options
//
int       opt_waitstates  = 0;
int       opt_standby     = 0;
int       opt_timeout     = 10;
// int       opt_device      = 0;
int       opt_partition[4] = {0, -1, -1, -1};
bool      opt_showhelp    = false;
bool      opt_readonly    = false;
bool      opt_verbose     = false;
bool      opt_acoustic    = false;
unsigned  opt_iobase      = 0x100;

CmdOpt  cmdoptions[] = {
  {"/s",  "%d", &opt_standby,    "nnn  stand-by time in seconds" },
  {"/t",  "%d", &opt_timeout,    "nn   time out in seconds"  },
//  {"/d",  "%d", &opt_device,     "n    device (0=master, 1=slave)" },
  {"/p",  "%d", &opt_partition[0],  "n    first partition number" },
  {"/p2",  "%d", &opt_partition[1],  "n    second partition number" },
  {"/p3",  "%d", &opt_partition[2],  "n    third partition number" },
  {"/p4",  "%d", &opt_partition[3],  "n    fourth partition number" },
  {"/r",     0, &opt_readonly,   "     read-only mode" },
  {"/v",     0, &opt_verbose,    "     verbose mode" },
  {"/b",  "%x", &opt_iobase,     "xxx  IDE port base address" },
  {"/w",  "%d", &opt_waitstates, "nn   number of waitstates [0-15]" },
  {"/a",     0, &opt_acoustic,   "     low noise acoustic management" },
  {"/?",     0, &opt_showhelp,   "     show help" },
  {0,0,0,0}
};


static  Partition  partitions[MAX_PARTITIONS];
static  int        numPartitions;

// Some global variables for the ExtDisk handler
//
static unsigned long  EXTDISK_LBAOffset;
static volatile int   EXTDISK_StandbyTimer;

static void readPartitionTable(unsigned long lba)
{
  // caution: remember that this is static!
  static     PartSector psec;
  unsigned   c, h, s;
  LBA_to_PCHS(lba, &c, &h, &s);
  if (opt_verbose)
    printf("reading partition table at %ld..\n", lba);

  if (!IDE_Read(c, h, s, 1, (char*)&psec) || psec.magic != 0xAA55) {
    printf("\nERROR: invalid partition table at %ld.\n", lba);
    return;
  }
  // Read partition table
  //
  unsigned long nextPart = 0;
  for (int i=0; i<4; i++) {
    if (psec.part[i].type != 0x00 &&
        psec.part[i].type != 0x05 )
    {
      if (numPartitions >= MAX_PARTITIONS)
        return;
      Partition *p = &partitions[numPartitions];
      p->type  = psec.part[i].type;
      p->start = lba + psec.part[i].offset;
      p->size  = psec.part[i].size;
      numPartitions++;
    }
    if (psec.part[i].type == 0x05)
      nextPart = lba + psec.part[i].offset;
  }
  // Recursively handle extended partitions
  //
  if (nextPart)
    readPartitionTable(nextPart);
}

int main(int argc, char *argv[])
{
  printf( "External IDE disk driver v0.99b\n" );
  printf( "(c)2001 thomas.kindler@gmx.de\n" );
  printf( "Filesystem ported from Salvatore Isaja's FreeDOS32 FAT Driver\n");
  printf( "by Kai Gossner, xyphro@gmx.net\n" );

  // Parse command line options
  //
  if (!GetOptions(argc, argv, cmdoptions) || opt_showhelp) {
    ShowOptionHelp(cmdoptions);
    return -1;
  }

  // Initialize hardware
  //
  pfe_enable_pcs(opt_iobase >> 8);
  pfe_enable_bus(0xFFFF, 1);
  if ((opt_iobase >> 8) <= 3)
    AMD_SetPcsWaitStates(opt_waitstates);

  if (opt_verbose)
    printf( "using interface at 0x%x, %d waitstates.\n\n",
            opt_iobase, opt_waitstates);

  // Identify disk, set LBA->PCHS translation,
  // and display some information
  //
  DeviceInfo  drive;

  if (opt_verbose)
    printf("identifying disk..\n");
  if (!IDE_Identify(&drive))  {
    printf("\nERROR: identify failed.\n");
    return -1;
  }

  SetPCHSTranslation( drive.heads, drive.sectors );
  if (opt_verbose) {
    printf("  drive id: %04Xh, model: \"%s\",\n", drive.id, drive.model);
    printf("  firmware rev. %s, serial #%s,\n", drive.firmware, drive.serial);
    printf("  %d cylinders, %d heads, %d sectors,\n",
           drive.cylinders, drive.heads, drive.sectors);
    printf("  %lu total sectors x 512 bytes, \n",  drive.totalSectors);
    printf("  %lu kb (%lu mb).\n\n", drive.totalKBytes, drive.totalKBytes >> 10);
  }

/*  printf("setting power up in standby mode..\n");
  IDE_SetFeatures( FEAT_EN_POWER_UP_IN_STANDBY, 0,0,0,0 );
*/
  if (opt_acoustic) {
    if (opt_verbose) {
      printf("setting acoustic management to low noise..\n");
    }
    IDE_SetFeatures( FEAT_EN_ACOUSTIC_MANAGE, 0,0,0, 0x80 );
  }

  // Recursively read partition tables
  //
  readPartitionTable(0);
  if (opt_verbose) {
    for (int i=0; i<numPartitions; i++) {
      printf( "  #%d type %02Xh at LBA %ld, %ld kb (%ld mb)\n",
              i, partitions[i].type, partitions[i].start,
              partitions[i].size>>1, partitions[i].size >> 11 );
    }
    printf("\n");
  }

  for (int i=0; i< MAX_DEVICE_COUNT; i++)
    devices[i] = 0;
  devicecount = 0;

  // Check if partition is valid
  //
  for (int i=0; i < 4; i++) {
    if (opt_partition[i] >= numPartitions) {
      printf( "ERROR: partition #%d doesn't exist.\n",
              opt_partition);
      return -1;
    }
  }


//  printf("test: %s\n", (char*) MK_FP(Seg,Off));
  for (int i=0; i < 4; i++) {
    if (opt_partition[i] > -1) {
      fat_mount(&partitions[opt_partition[i]]);
      printf("\nMounted partition #%d as Drive %s\n", opt_partition[i], devices[0]->DevName);
    }
  }

  InstallIntHandler();

  unsigned long t;
//  EXTDISK_StandbyTimer = opt_standby;
  while ( !driver_unmount) {
    RTX_Sleep_Time(1000);
    if (EXTDISK_StandbyTimer==0 && opt_standby!=0) {
//      IDE_Standby();
      EXTDISK_StandbyTimer = -1;
    } else {
      EXTDISK_StandbyTimer--;
    }
  }
  printf("Unmounted.\n");
  UnInstallIntHandler();
  fat_unmount();
}
